#!/usr/bin/env python3

"""
Iterates through any dataset, and prints/dump some information.
This can also be used to collect statistics over the data like mean/variance.
"""

from __future__ import annotations

import os
import sys
import time
import typing

import _setup_returnn_env  # noqa
from returnn import __main__ as rnn
from returnn.log import log
import argparse
import numpy
from returnn.datasets import init_dataset, Dataset
from returnn.util.basic import Stats, hms, hms_fraction, pretty_print
from returnn.util import basic as util


dataset = None  # type: typing.Optional[Dataset]


def plot(m):
    """
    :param numpy.ndarray m:
    """
    print("Plotting matrix of shape %s." % (m.shape,))
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    from matplotlib.pyplot import matshow, show

    matshow(m.transpose())
    show()


def dump_dataset(options):
    """
    :param options: argparse.Namespace
    """
    print("Epoch: %i" % options.epoch, file=log.v3)
    seq_list = None
    if options.seqtags:
        seq_list = options.seqtags.split(",")
    dataset.init_seq_order(epoch=options.epoch, seq_list=seq_list)
    print("Dataset keys:", dataset.get_data_keys(), file=log.v3)
    print("Dataset target keys:", dataset.get_target_list(), file=log.v3)
    assert options.key in dataset.get_data_keys()

    if options.get_num_seqs:
        print("Get num seqs.")
        print("estimated_num_seqs: %r" % dataset.estimated_num_seqs)
        try:
            print("num_seqs: %r" % dataset.num_seqs)
        except Exception as exc:
            print("num_seqs exception %r, which is valid, so we count." % exc)
            seq_idx = 0
            if dataset.get_target_list():
                default_target = dataset.get_target_list()[0]
            else:
                default_target = None
            while dataset.is_less_than_num_seqs(seq_idx):
                dataset.load_seqs(seq_idx, seq_idx + 1)
                if seq_idx % 10000 == 0:
                    if default_target:
                        targets = dataset.get_data(seq_idx, default_target)
                        postfix = " (targets = %r...)" % (targets[:10],)
                    else:
                        postfix = ""
                    print("%i ...%s" % (seq_idx, postfix))
                seq_idx += 1
            print("accumulated num seqs: %i" % seq_idx)
        print("Done.")
        return

    dump_file = None
    if options.type == "numpy":
        print("Dump files: %r*%r" % (options.dump_prefix, options.dump_postfix), file=log.v3)
    elif options.type == "stdout":
        print("Dump to stdout", file=log.v3)
        if options.stdout_limit is not None:
            util.set_pretty_print_default_limit(options.stdout_limit)
            numpy.set_printoptions(
                threshold=sys.maxsize if options.stdout_limit == float("inf") else int(options.stdout_limit)
            )
        if options.stdout_as_bytes:
            util.set_pretty_print_as_bytes(options.stdout_as_bytes)
    elif options.type == "print_tag":
        print("Dump seq tag to stdout", file=log.v3)
    elif options.type == "dump_tag":
        dump_file = open("%sseq-tags.txt" % options.dump_prefix, "w")
        print("Dump seq tag to file: %s" % (dump_file.name,), file=log.v3)
    elif options.type == "dump_seq_len":
        dump_file = open("%sseq-lens.txt" % options.dump_prefix, "w")
        print("Dump seq lens to file: %s" % (dump_file.name,), file=log.v3)
        dump_file.write("{\n")
    elif options.type == "print_shape":
        print("Dump shape to stdout", file=log.v3)
    elif options.type == "plot":
        print("Plot.", file=log.v3)
    elif options.type == "interactive":
        print("Interactive debug shell.", file=log.v3)
    elif options.type == "null":
        if options.dump_stats:
            print("No dump (except stats).")
        else:
            print("No dump.")
    else:
        raise Exception("unknown dump option type %r" % options.type)

    start_time = time.time()
    stats = Stats() if (options.stats or options.dump_stats) else None
    seq_len_stats = {key: Stats() for key in dataset.get_data_keys()}
    seq_idx = options.startseq
    if options.endseq < 0:
        options.endseq = float("inf")
    while dataset.is_less_than_num_seqs(seq_idx) and seq_idx <= options.endseq:
        dataset.load_seqs(seq_idx, seq_idx + 1)
        complete_frac = dataset.get_complete_frac(seq_idx)
        start_elapsed = time.time() - start_time
        try:
            num_seqs_s = str(dataset.num_seqs)
        except NotImplementedError:
            try:
                num_seqs_s = "~%i" % dataset.estimated_num_seqs
            except TypeError:  # a number is required, not NoneType
                num_seqs_s = "?"
        progress_prefix = "%i/%s" % (seq_idx, num_seqs_s)
        progress = "%s (%.02f%%)" % (progress_prefix, complete_frac * 100)
        data = None
        if complete_frac > 0:
            total_time_estimated = start_elapsed / complete_frac
            remaining_estimated = total_time_estimated - start_elapsed
            progress += " (%s)" % hms(remaining_estimated)
        if options.type == "print_tag":
            print("seq %s tag:" % (progress if log.verbose[2] else progress_prefix), dataset.get_tag(seq_idx))
        elif options.type == "dump_tag":
            print("seq %s tag:" % (progress if log.verbose[2] else progress_prefix), dataset.get_tag(seq_idx))
            dump_file.write("%s\n" % dataset.get_tag(seq_idx))
        elif options.type == "dump_seq_len":
            seq_len = dataset.get_seq_length(seq_idx)[options.key]
            print(
                "seq %s tag:" % (progress if log.verbose[2] else progress_prefix),
                dataset.get_tag(seq_idx),
                "%r len:" % options.key,
                seq_len,
            )
            dump_file.write("%r: %r,\n" % (dataset.get_tag(seq_idx), seq_len))
        else:
            data = dataset.get_data(seq_idx, options.key)
            if options.type == "numpy":
                numpy.savetxt("%s%i.data%s" % (options.dump_prefix, seq_idx, options.dump_postfix), data)
            elif options.type == "stdout":
                print("seq %s tag:" % progress, dataset.get_tag(seq_idx))
                print("seq %s data:" % progress, pretty_print(data))
            elif options.type == "print_shape":
                print("seq %s data shape:" % progress, data.shape)
            elif options.type == "plot":
                plot(data)
            for target in dataset.get_target_list():
                targets = dataset.get_data(seq_idx, target)
                if options.type == "numpy":
                    numpy.savetxt(
                        "%s%i.targets.%s%s" % (options.dump_prefix, seq_idx, target, options.dump_postfix),
                        targets,
                        fmt="%i",
                    )
                elif options.type == "stdout":
                    extra = ""
                    if target in dataset.labels and len(dataset.labels[target]) > 1:
                        assert dataset.can_serialize_data(target)
                        extra += " (%r)" % dataset.serialize_data(key=target, data=targets)
                    print("seq %i target %r: %s%s" % (seq_idx, target, pretty_print(targets), extra))
                elif options.type == "print_shape":
                    print("seq %i target %r shape:" % (seq_idx, target), targets.shape)
            if options.type == "interactive":
                from returnn.util.debug import debug_shell

                debug_shell(locals())
        seq_len = dataset.get_seq_length(seq_idx)
        for key in dataset.get_data_keys():
            seq_len_stats[key].collect([seq_len[key]])
        if stats:
            stats.collect(data)
        if options.type == "null":
            util.progress_bar_with_time(complete_frac, prefix=progress_prefix)

        seq_idx += 1

    print(
        "Done. Total time %s. More seqs which we did not dumped: %s"
        % (hms_fraction(time.time() - start_time), dataset.is_less_than_num_seqs(seq_idx)),
        file=log.v2,
    )
    for key in dataset.get_data_keys():
        seq_len_stats[key].dump(stream_prefix="Seq-length %r " % key, stream=log.v2)
    if stats:
        stats.dump(output_file_prefix=options.dump_stats, stream_prefix="Data %r " % options.key, stream=log.v1)
    if options.type == "dump_seq_len":
        dump_file.write("}\n")
    if dump_file:
        print("Dumped to file:", dump_file.name, file=log.v2)
        dump_file.close()


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_str, config_dataset, verbosity):
    """
    :param str config_str: either filename to config-file, or dict for dataset
    :param str|None config_dataset:
    :param int verbosity:
    """
    global dataset
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    dataset_dict = None
    config_filename = None
    if config_str.strip().startswith("{"):
        print("Using dataset %s." % config_str)
        dataset_dict = eval(config_str.strip())
    elif config_str.endswith(".hdf"):
        dataset_dict = {"class": "HDFDataset", "files": [config_str]}
        print("Using dataset %r." % dataset_dict)
        assert os.path.exists(config_str)
    else:
        config_filename = config_str
        print("Using config file %r." % config_filename)
        assert os.path.exists(config_filename)
    rnn.init_config(config_filename=config_filename, default_config={"cache_size": "0"})
    global config
    config = rnn.config
    config.set("log", None)
    config.set("log_verbosity", verbosity)
    config.set("torch_distributed", None)
    config.set("use_horovod", None)
    if dataset_dict:
        assert not config_dataset
        dataset = init_dataset(dataset_dict)
    elif config_dataset and config_dataset != "train":
        print("Use dataset %r from config." % config_dataset)
        dataset = init_dataset("config:%s" % config_dataset)
    else:
        print("Use train dataset from config.")
        assert config.value("train", None)
        dataset = init_dataset("config:train")
    rnn.init_log()
    print("Returnn dump-dataset starting up.", file=log.v2)
    rnn.returnn_greeting()
    rnn.init_faulthandler()
    print("Dataset:", file=log.v2)
    print("  input:", dataset.num_inputs, "x", dataset.window, file=log.v2)
    print("  output:", dataset.num_outputs, file=log.v2)
    print(" ", dataset.len_info() or "no info", file=log.v2)


def main():
    """
    Main entry.
    """
    argparser = argparse.ArgumentParser(description="Dump something from dataset.")
    argparser.add_argument("returnn_config", help="either filename to config-file, or dict for dataset")
    argparser.add_argument("--dataset", help="if given the config, specifies the dataset. e.g. 'dev'")
    argparser.add_argument("--epoch", type=int, default=1)
    argparser.add_argument("--startseq", type=int, default=0, help="start seq idx (inclusive) (default: 0)")
    argparser.add_argument("--endseq", type=int, default=None, help="end seq idx (inclusive) or -1 (default: 10)")
    argparser.add_argument("--seqtags", type=str, default=None, help="comma-separated list of seq-tags to dump")
    argparser.add_argument("--get_num_seqs", action="store_true")
    argparser.add_argument("--type", default="stdout", help="'numpy', 'stdout', 'plot', 'null' (default 'stdout')")
    argparser.add_argument("--stdout_limit", type=float, default=None, help="e.g. inf to disable")
    argparser.add_argument("--stdout_as_bytes", action="store_true")
    argparser.add_argument("--verbosity", type=int, default=4, help="overwrites log_verbosity (default: 4)")
    argparser.add_argument("--dump_prefix", default="/tmp/returnn.dump-dataset.")
    argparser.add_argument("--dump_postfix", default=".txt.gz")
    argparser.add_argument("--key", default="data", help="data-key, e.g. 'data' or 'classes'. (default: 'data')")
    argparser.add_argument("--stats", action="store_true", help="calculate mean/stddev stats")
    argparser.add_argument("--dump_stats", help="file-prefix to dump stats to")
    args = argparser.parse_args()
    if args.endseq is None:
        args.endseq = 10 if not args.seqtags else -1
    init(config_str=args.returnn_config, config_dataset=args.dataset, verbosity=args.verbosity)
    try:
        dump_dataset(args)
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
        sys.exit(1)
    finally:
        rnn.finalize()


if __name__ == "__main__":
    main()



"""
Start RETURNN training in a subprocess
and if it crashes, try to restart it under certain conditions.
"""

from __future__ import annotations

import os
import subprocess
import sys
import time

from returnn.config import Config
from returnn.util import basic as util
from returnn.engine.base import EngineBase


def maybe_start_train_proc_manager(*, config: Config):
    """
    Under certain conditions, start the train proc manager.
    """
    if os.environ.get("__RETURNN_PROC_MANAGED") == "1":
        print("Running in managed mode.")
        return
    if config.value("task", "train") != "train":
        return
    if config.value("start_epoch", "auto") != "auto":
        return

    # Not supported currently in distributed mode.
    if config.bool("use_horovod", False):
        return
    if config.typed_value("torch_distributed") is not None:
        return

    main_proc_manager(config=config)


def main_proc_manager(*, config: Config):
    """
    Start this in the parent.
    """
    os.environ["__RETURNN_PROC_MANAGED"] = "1"

    print("RETURNN train proc manager starting up, version %s" % util.describe_returnn_version())

    # Setup BackendEngine. This is that other utility functions know the right model file extension.
    # _select_rf_backend to avoid importing Torch/TF/anything here.
    util.BackendEngine.select_engine(config=config, _select_rf_backend=False)

    final_epoch = EngineBase.config_get_final_epoch(config)

    total_start_time = time.time()
    num_starts = 0
    last_model_epoch = None
    return_code = None
    while True:
        models = EngineBase.get_existing_models(config, for_training=True)
        cur_model_epoch = max(models) if models else 0
        print("Most recent trained model epoch:", cur_model_epoch, "file:", models.get(cur_model_epoch))

        if cur_model_epoch == final_epoch:
            if return_code is None:
                return_code = 0
                print("Already finished, nothing to do")
            elif return_code == 0:
                print("Finished training")
            else:
                print("Finished training, but got some error?")
            break

        if last_model_epoch is not None:
            print("Most recent trained model epoch before RETURNN run:", last_model_epoch)
            assert last_model_epoch <= cur_model_epoch
            print(f"-> trained successfully {cur_model_epoch - last_model_epoch} epoch(s)")
            if cur_model_epoch == last_model_epoch:
                print("-> break")
                break
            print("Try again, restart RETURNN...")
        else:
            print("Run RETURNN...")

        sys.stdout.flush()
        last_model_epoch = cur_model_epoch
        num_starts += 1
        start_time = time.time()
        args = [sys.executable] + sys.argv
        print(f"Run {args}")
        proc = subprocess.Popen(args)
        proc.wait()
        print("RETURNN runtime:", util.hms(time.time() - start_time))
        print("RETURNN return code:", proc.returncode)
        return_code = proc.returncode

    print("Total RETURNN num starts:", num_starts)
    print("Total RETURNN runtime:", util.hms(time.time() - total_start_time))
    sys.exit(return_code if return_code is not None else -1)



"""
RETURNN frontend (returnn.frontend) tests
"""

from __future__ import annotations
from typing import Tuple
import numpy as np
import numpy.testing
import _setup_test_env  # noqa
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict, batch_dim
from rf_utils import run_model, tf_scope


def test_dot_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    key_dim = Dim(7, name="key")
    value_dim = Dim(13, name="value")
    extern_data = TensorDict(
        {
            "q": Tensor("q", [batch_dim, time_dim, key_dim], dtype="float32"),
            "k": Tensor("k", [batch_dim, time_dim, key_dim], dtype="float32"),
            "v": Tensor("v", [batch_dim, time_dim, value_dim], dtype="float32", feature_dim_axis=2),
        }
    )

    class _Net(rf.Module):
        def __call__(self, q: Tensor, k: Tensor, v: Tensor) -> Tensor:
            kv_axis = Dim(None, name=f"kv-axis")
            k, _ = rf.replace_dim(k, in_dim=time_dim, out_dim=kv_axis)
            v, _ = rf.replace_dim(v, in_dim=time_dim, out_dim=kv_axis)
            return rf.dot_attention(q, k, v, axis=kv_axis, key_dim=key_dim)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(q=extern_data["q"], k=extern_data["k"], v=extern_data["v"])
        out.mark_as_default_output(shape=(batch_dim, time_dim, value_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_self_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.SelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""
            return self.self_att(x, axis=axis)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_causal_self_attention():
    from returnn.tensor import single_step_dim

    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(7, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.CausalSelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""

            def _body(_x: Tensor, _state: rf.State) -> Tuple[Tensor, rf.State]:
                _y, _state.self_att = self.self_att(_x, axis=single_step_dim, state=_state.self_att)
                return _y, _state

            y, _, _ = rf.scan(
                spatial_dim=axis,
                xs=x,
                body=_body,
                ys=Tensor("y", dims=[batch_dim, self.out_dim], dtype="float32"),
                initial=rf.State(self_att=self.self_att.default_initial_state(batch_dims=[batch_dim])),
            )
            return y

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(
        extern_data,
        lambda *, epoch, step: _Net(),
        _forward_step,
        # TF needs TensorArray unstack, not implemented yet
        test_tensorflow=False,
    )


def test_relative_positional_encoding():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(8, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __call__(self, x: Tensor, *, axis: Dim) -> Tuple[Tensor, Dim]:
            x, dim = rf.relative_positional_encoding(
                key_value_spatial_dim=axis, query_spatial_dim=axis, feat_dim=in_dim
            )
            return x, dim

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out, dim = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(dim, in_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_rel_pos_self_attention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    in_dim = Dim(8, name="in")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, in_dim], dtype="float32"),
        }
    )

    class _Net(rf.Module):
        def __init__(self):
            super().__init__()
            self.self_att = rf.RelPosSelfAttention(
                in_dim=in_dim,
                proj_dim=Dim(5, name="out"),
                key_dim_total=Dim(21, name="key-dim-total"),
                value_dim_total=Dim(33, name="value-dim-total"),
                num_heads=3,
            )
            self.out_dim = self.self_att.out_dim

        def __call__(self, x: Tensor, *, axis: Dim) -> Tensor:
            """forward"""
            return self.self_att(x, axis=axis)

    # noinspection PyShadowingNames
    def _forward_step(*, model: _Net, extern_data: TensorDict):
        out = model(extern_data["data"], axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, model.out_dim))

    run_model(extern_data, lambda *, epoch, step: _Net(), _forward_step)


def test_sinusoidal_positional_encoding():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(8, name="feat")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, feat_dim], dtype="float32"),
        }
    )

    def _forward_step(**_kwargs):
        out = rf.sinusoidal_positional_encoding(spatial_dim=time_dim, feat_dim=feat_dim)
        out.mark_as_default_output(shape=(time_dim, feat_dim))

    res = run_model(extern_data, lambda *, epoch, step: rf.Module(), _forward_step)

    from returnn.tf.util import basic as tf_util

    with tf_scope() as session:
        tf_ref = tf_util.get_positional_encoding(
            num_channels=feat_dim.dimension, length=res.data["output"].raw_tensor.shape[0]
        )
        tf_ref_v = session.run(tf_ref)

    np.testing.assert_almost_equal(res.data["output"].raw_tensor, tf_ref_v, decimal=5)


def test_CausalSelfAttention():
    time_dim = Dim(Tensor("time", [batch_dim], dtype="int32"))
    feat_dim = Dim(8, name="feat")
    key_dim = Dim(6, name="key")
    value_dim = Dim(10, name="value")
    extern_data = TensorDict(
        {
            "data": Tensor("data", [batch_dim, time_dim, feat_dim], dtype="float32"),
        }
    )

    def _forward_step(*, model: rf.CausalSelfAttention, extern_data: TensorDict):
        data = extern_data["data"]
        data.mark_as_output("data", shape=[batch_dim, time_dim, feat_dim])
        time_dim.dyn_size_ext.mark_as_output("seq_len", shape=[batch_dim])
        out, _ = model(data, axis=time_dim)
        out.mark_as_default_output(shape=(batch_dim, time_dim, value_dim))
        model.qkv.weight.mark_as_output("qkv_weight", shape=[feat_dim, 2 * key_dim + value_dim])

    res = run_model(
        extern_data,
        lambda *, epoch, step: rf.CausalSelfAttention(
            in_dim=feat_dim,
            proj_dim=None,
            key_dim_total=key_dim,
            value_dim_total=value_dim,
            num_heads=2,
            with_bias=False,
        ),
        _forward_step,
        # Some problem with dimension tags currently in the TF-layers-dict backend...
        # Anyway, we compare to the TF SelfAttentionLayer with attention_left_only=True below.
        test_tensorflow=False,
    )

    extern_data.reset_content()

    with tf_scope() as session:
        from returnn.tf.network import TFNetwork, ExternData

        net_dict = {
            "self_att": {
                "class": "self_attention",
                "from": "data",
                "num_heads": 2,
                "total_key_dim": key_dim.dimension,
                "attention_left_only": True,
                "out_dim": value_dim,
                "is_output_layer": True,
            }
        }
        net = TFNetwork(
            extern_data=ExternData(
                {
                    "data": {
                        "dims": [batch_dim, time_dim, feat_dim],
                        "time_dim_axis": 1,
                        "feature_dim_axis": 2,
                        "dtype": "float32",
                        "version": 1,
                    }
                }
            )
        )
        net.construct_from_dict(net_dict)
        layer = net.get_default_output_layer()
        layer.params["QKV"].load(res.data["qkv_weight"].raw_tensor, session=session)
        out = layer.output.copy_transpose([batch_dim, time_dim, value_dim]).copy_masked(0.0)

        out_tf_v = session.run(
            out.raw_tensor,
            feed_dict={
                net.extern_data.data["data"].placeholder: res.data["data"].raw_tensor,
                net.extern_data.data["data"].dims[1].dyn_size_ext.raw_tensor: res.data["seq_len"].raw_tensor,
            },
        )
        numpy.testing.assert_almost_equal(res.data["output"].raw_tensor, out_tf_v, decimal=5)



"""
File cache.

Copies files from a remote filesystem (e.g. NFS)
to a local filesystem (e.g. /var/tmp) to speed up access.

See https://github.com/rwth-i6/returnn/issues/1519 for initial discussion.

Main class is :class:`FileCache`.
"""

from typing import Any, Collection, List, Tuple
import os
import time
import shutil
from dataclasses import dataclass
from collections import defaultdict
from contextlib import contextmanager
from threading import Thread, Event
from .basic import expand_env_vars, LockFile, human_bytes_size


__all__ = ["FileCache", "CachedFile"]


class FileCache:
    """
    File cache.

    Copies files from a remote filesystem (e.g. NFS)
    to a local filesystem (e.g. ``/var/tmp``) to speed up access.

    Some assumptions we depend on:

    - When a cached file is available, and its size matches the original file,
      we can use it.
      (We might want to extend this logic later, e.g. by storing and checking the original source mtime.
       But for simplicity, we don't do this for now.)
    - We will update the cached file mtime frequently (every second) via a background thread
      of used cached files, to mark that they are used.
      (We would maybe want to use atime, but we don't expect that atime can be relied on.)
      Note that updating mtime might influence the behavior of some external tools.
      (If this becomes a problem at some point, instead of abusing mtime this way,
       we might store the most recent usage time e.g. in a separate file.)
    - :func:`os.utime` will update mtime, and mtime is somewhat accurate (up to 10 secs maybe),
      mtime compares to time.time().
    - :func:`shutil.disk_usage` can be relied on.

    See https://github.com/rwth-i6/returnn/issues/1519 for initial discussion.
    """

    def __init__(
        self,
        *,
        cache_directory: str = "$TMPDIR/$USER/returnn/file_cache",
        cleanup_files_always_older_than_days: float = 31.0,
        cleanup_files_wanted_older_than_days: float = 7.0,
        cleanup_disk_usage_wanted_free_ratio: float = 0.2,  # try to free at least 20% disk space
    ):
        """
        :param cache_directory: directory where to cache files.
            Uses :func:`expand_env_vars` to expand environment variables.
        :param cleanup_files_always_older_than_days: always cleanup files older than this.
        :param cleanup_files_wanted_older_than_days: if cleanup_disk_usage_wanted_free_ratio not reached,
            cleanup files older than this.
        :param cleanup_disk_usage_wanted_free_ratio: try to free at least this ratio of disk space.
        """
        self.cache_directory = expand_env_vars(cache_directory)
        self._cleanup_files_always_older_than_days = cleanup_files_always_older_than_days
        self._cleanup_files_wanted_older_than_days = cleanup_files_wanted_older_than_days
        self._cleanup_disk_usage_wanted_free_ratio = cleanup_disk_usage_wanted_free_ratio
        self._touch_files_thread = _TouchFilesThread(cache_base_dir=self.cache_directory)
        self._touch_files_thread.start()
        self._recent_full_cleanup_time = float("-inf")

    # Note on lock_timeout: It will check whether a potentially existing lock file is older than this timeout,
    # and if so, then it would delete the existing lock file, assuming it is from a crashed previous run.
    # We are always keeping the lock file mtime updated via the _touch_files_thread (every second),
    # so it should never be older than this timeout.
    # If there is a Python exception anywhere here, we will always properly release the lock.
    # Only if the process dies (e.g. killed, segfault or so), the lock file might be left-over,
    # and another process might need to wait for this timeout.
    # We don't expect that this must be configured, so let's just use a reasonable default.
    # This should be more than the _touch_files_thread interval (1 sec).
    _lock_timeout = 20

    def __del__(self):
        self._touch_files_thread.stop.set()

    def get_file(self, src_filename: str) -> str:
        """
        Get cached file.
        This will copy the file to the cache directory if it is not already there.
        This will also make sure that the file is not removed from the cache directory
        via the _touch_files_thread
        until you call :func:`release_file`.

        :param src_filename: source file to copy (if it is not already in the cache).
        :return: cached file path (in the cache directory)
        """
        dst_filename = self._get_dst_filename(src_filename)
        self._copy_file_if_needed(src_filename, dst_filename)
        self._touch_files_thread.files_extend([dst_filename])
        return dst_filename

    def release_files(self, filenames: Collection[str]):
        """
        Release cached files.
        This just says that we are not using the files anymore for now.
        They will be kept in the cache directory for now,
        and might be removed when the cache directory is cleaned up.

        :param filenames: files to release (paths in the cache directory)
        """
        self._touch_files_thread.files_remove(filenames)

    def cleanup(self, *, need_at_least_free_space_size: int = 0):
        """
        Cleanup cache directory.
        """
        if not os.path.exists(self.cache_directory):
            return
        disk_usage = shutil.disk_usage(self.cache_directory)
        want_free_space_size = max(
            need_at_least_free_space_size, int(self._cleanup_disk_usage_wanted_free_ratio * disk_usage.total)
        )
        # If we have enough free space, and we did a full cleanup recently, we don't need to do anything.
        if want_free_space_size <= disk_usage.free and time.monotonic() - self._recent_full_cleanup_time < 60 * 10:
            return
        # Do a full cleanup, i.e. iterate through all files in cache directory and check their mtime.
        # Get current time now, so that cur_time - mtime is pessimistic,
        # and does not count the time for the cleanup itself.
        cur_time = time.time()
        all_files = []  # mtime, neg size (better for sorting), filename
        for root, dirs, files in os.walk(self.cache_directory):
            for rel_fn in files:
                fn = root + "/" + rel_fn
                try:
                    f_stat = os.stat(fn)
                except Exception as exc:
                    print(f"FileCache: Error while stat {fn}: {type(exc).__name__}: {exc}")
                    continue
                else:
                    all_files.append((f_stat.st_mtime, -f_stat.st_blocks * 512, fn))
        all_files.sort()
        cur_expected_free = disk_usage.free
        reached_more_recent_files = False
        cur_used_time_threshold = self._lock_timeout * 0.5  # Used files mtime should be updated every second.
        total_cache_files_size = sum(-neg_size for _, neg_size, _ in all_files)
        total_cur_used_cache_files_size = sum(
            -neg_size for mtime, neg_size, fn in all_files if cur_time - mtime <= cur_used_time_threshold
        )
        report_size_str = (
            f"Total size cached files: {human_bytes_size(total_cache_files_size)},"
            f" currently used: {human_bytes_size(total_cur_used_cache_files_size)}"
        )
        for mtime, neg_size, fn in all_files:
            size = -neg_size
            delete_reason = None
            if cur_time - mtime > self._cleanup_files_always_older_than_days * 60 * 60 * 24:
                delete_reason = f"File is {(cur_time - mtime) / 60 / 60 / 24} days old"
            else:
                reached_more_recent_files = True
            if not delete_reason and need_at_least_free_space_size > cur_expected_free:
                # Still must delete some files.
                if cur_time - mtime > cur_used_time_threshold:
                    delete_reason = f"Still need more space, file is {cur_time - mtime} secs old"
                else:
                    raise Exception(
                        f"We cannot free enough space on {self.cache_directory}.\n"
                        f"Needed: {human_bytes_size(need_at_least_free_space_size)},\n"
                        f"currently available: {human_bytes_size(cur_expected_free)},\n"
                        f"oldest file is still too recent: {fn}.\n"
                        f"{report_size_str}"
                    )
            if not delete_reason and want_free_space_size > cur_expected_free:
                if cur_time - mtime > self._cleanup_files_wanted_older_than_days * 60 * 60 * 24:
                    delete_reason = f"Still want more space, file is {(cur_time - mtime) / 60} min old"
                else:
                    # All further files are even more recent, so we would neither cleanup them,
                    # so we can also just stop now.
                    break

            if delete_reason:
                cur_expected_free += size
                print(
                    f"FileCache: Delete file {fn}, size {human_bytes_size(size)}. {delete_reason}."
                    f" After deletion, have {human_bytes_size(cur_expected_free)} free space."
                )
                try:
                    os.remove(fn)
                except Exception as exc:
                    print(f"FileCache: Error while removing {fn}: {type(exc).__name__}: {exc}")
                    cur_expected_free -= size

            if reached_more_recent_files and want_free_space_size <= cur_expected_free:
                # Have enough free space now.
                break

        if need_at_least_free_space_size > cur_expected_free:
            raise Exception(
                f"We cannot free enough space on {self.cache_directory}.\n"
                f"Needed: {human_bytes_size(need_at_least_free_space_size)},\n"
                f"currently available: {human_bytes_size(cur_expected_free)}.\n"
                f"{report_size_str}"
            )

        # Cleanup empty dirs.
        for root, dirs, files in os.walk(self.cache_directory, topdown=False):
            if files:
                continue
            try:
                if cur_time - os.stat(root).st_mtime <= cur_used_time_threshold:  # still in use?
                    continue
            except Exception as exc:
                print(f"FileCache: Error while stat dir {root}: {type(exc).__name__}: {exc}")
                continue
            try:
                # Recheck existence of dirs, because they might have been deleted by us.
                if any(os.path.exists(root + "/" + d) for d in dirs):
                    continue
            except Exception as exc:
                print(f"FileCache: Error while checking sub dirs in {root}: {type(exc).__name__}: {exc}")
                continue
            try:
                # We can delete this empty dir.
                print(f"FileCache: Remove empty dir {root}")
                os.rmdir(root)
            except Exception as exc:
                print(f"FileCache: Error while removing empty dir {root}: {type(exc).__name__}: {exc}")

        self._recent_full_cleanup_time = time.monotonic()

    def handle_cached_files_in_config(self, config: Any) -> Tuple[Any, List[str]]:
        """
        :param config: some config, e.g. dict, or any nested structure
        :return: modified config, where all :class:`CachedFile` instances are replaced by the cached file path,
            and the list of cached files which are used.
        """
        import tree

        res_files = []

        def _handle_value(value):
            if isinstance(value, CachedFile):
                res = self.get_file(value.filename)
                res_files.append(res)
                return res
            return value

        return tree.map_structure(_handle_value, config), res_files

    def _get_dst_filename(self, src_filename: str) -> str:
        """
        Get the destination filename in the cache directory.
        """
        assert src_filename.startswith("/")
        return self.cache_directory + src_filename

    def _copy_file_if_needed(self, src_filename: str, dst_filename: str):
        """
        Copy the file to the cache directory.
        """
        if self._check_existing_copied_file_maybe_cleanup(src_filename, dst_filename):
            os.utime(dst_filename, None)  # touch
            return

        # Make sure we have enough disk space.
        self.cleanup(need_at_least_free_space_size=os.stat(src_filename).st_size)

        print(f"FileCache: Copy file {src_filename} to cache")

        # Create dirs.
        dst_dir = os.path.dirname(dst_filename)
        os.makedirs(dst_dir, exist_ok=True)

        # Copy the file, while holding a lock. See comment on lock_timeout above.
        with LockFile(
            directory=dst_dir, name=os.path.basename(dst_filename) + ".lock", lock_timeout=self._lock_timeout
        ) as lock:
            # Maybe it was copied in the meantime, while waiting for the lock.
            if os.path.exists(dst_filename):
                return

            dst_tmp_filename = dst_filename + ".copy"
            if os.path.exists(dst_tmp_filename):
                # The minimum age should be at least the lock_timeout.
                # (But leave a bit of room for variance in timing in the sanity check below.)
                dst_tmp_file_age = time.time() - os.stat(dst_tmp_filename).st_mtime
                assert dst_tmp_file_age > self._lock_timeout * 0.8, (
                    f"FileCache: Expected left-over temp copy file {dst_tmp_filename}"
                    f" from crashed previous copy attempt"
                    f" to be older than {self._lock_timeout * 0.8}s but it is {dst_tmp_file_age} seconds old"
                )

            with self._touch_files_thread.files_added_context([dst_dir, lock.lockfile]):
                shutil.copyfile(src_filename, dst_tmp_filename)
                os.rename(dst_tmp_filename, dst_filename)

    @staticmethod
    def _check_existing_copied_file_maybe_cleanup(src_filename: str, dst_filename: str) -> bool:
        """
        Check if the file is in the cache directory.
        """
        if not os.path.exists(dst_filename):
            return False
        src_stat = os.stat(src_filename)
        dst_stat = os.stat(dst_filename)
        if src_stat.st_size != dst_stat.st_size:
            os.remove(dst_filename)
            return False
        return True


@dataclass
class CachedFile:
    """
    Represents some file to be cached in a user config.
    See :func:`FileCache.handle_cached_files_in_config`.
    """

    filename: str  # original filename


class _TouchFilesThread(Thread):
    def __init__(self, *, interval: float = 1.0, cache_base_dir: str):
        super().__init__(daemon=True)
        self.stop = Event()
        self.files = defaultdict(int)  # usage counter
        self.interval = interval
        self.cache_base_dir = cache_base_dir

    def run(self):
        """thread main loop"""
        while True:
            all_files = {}  # dict to have order deterministic
            for filename in self.files:
                all_files[filename] = True
                all_files.update({k: True for k in _all_parent_dirs(filename, base_dir=self.cache_base_dir)})
            for filename in all_files:
                os.utime(filename, None)
            if self.stop.wait(self.interval):
                return

    def files_extend(self, files: Collection[str]):
        """append"""
        assert isinstance(files, (list, set, tuple))
        for file in files:
            self.files[file] += 1

    def files_remove(self, files: Collection[str]):
        """remove"""
        if isinstance(files, str):
            files = [files]
        assert isinstance(files, (list, set, tuple))
        for filename in files:
            self.files[filename] -= 1
            if self.files[filename] <= 0:
                del self.files[filename]

    @contextmanager
    def files_added_context(self, files: Collection[str]):
        """temporarily add files, and remove them afterwards again."""
        self.files_extend(files)
        try:
            yield
        finally:
            self.files_remove(files)


def _all_parent_dirs(filename: str, *, base_dir: str) -> List[str]:
    assert filename.startswith(base_dir + "/")
    dirs = []
    while True:
        filename = os.path.dirname(filename)
        if filename == base_dir:
            break
        dirs.append(filename)
    return dirs



"""
:class:`ConcatFilesDataset `

https://github.com/rwth-i6/returnn/issues/1519
"""

from __future__ import annotations

from typing import Optional, Any, Dict, Tuple, List, Sequence, Collection, Callable, Union
import os
import sys
import numpy
from returnn.log import log
from returnn.util.basic import try_run
from returnn.util.multi_proc_non_daemonic_spawn import NonDaemonicSpawnContext
from returnn.config import SubProcCopyGlobalConfigPreInitFunc
from .basic import init_dataset, DatasetSeq
from .cached2 import CachedDataset2

# noinspection PyProtectedMember
from multiprocessing.connection import Connection as mpConnection

_mp = NonDaemonicSpawnContext(process_pre_init_func=SubProcCopyGlobalConfigPreInitFunc())


__all__ = ["ConcatFilesDataset"]

Filename = str
FileTree = Union[Filename, Tuple["FileTree", ...], Dict[Any, "FileTree"], List["FileTree"]]


class ConcatFilesDataset(CachedDataset2):
    """
    This is similar to :class:`ConcatDataset`, but instead of concatenating datasets,
    we distribute files over subepochs,
    and then create a sub dataset for every sub epoch
    for a given subset of the files
    via the given ``get_sub_epoch_dataset``.

    This scheme allows to shuffle over the files,
    which makes shuffling much more efficient over a large dataset
    at the cost of no longer shuffling over the full dataset in every subepoch.
    Instead, the quality of the shuffle depends on the number of files the dataset is
    split into -- the more files per subepoch, the better.

    Additionally, this scheme allows to prefetch and cache the upcoming needed files,
    e.g. copying them from a NFS to local disk,
    when the local disk can not store the whole dataset,
    and/or when the local disk would allow for faster access.
    For that, the user can use :class:`returnn.util.file_cache.CachedFile`
    inside the returned dict from ``get_sub_epoch_dataset``.

    It was also designed with multi-GPU training in mind,
    where each GPU should get a set of files to work on,
    and the sub epochs across the GPUs would not be exactly of the same length.
    Whenever one GPU finishes its sub epoch, all remaining data on the other GPUs is dropped.
    First, our scheme tries to make the length of the sub epochs as equally distributed as possible
    after random shuffling by using the size of the files as proxy for the length of the contents.
    Further, due to the random shuffling, there should not be any bias in the data distribution.
    Specifically, we don't want that some data might be visited more often than others
    (at least its expected value should be the same).

    Example usage::

        def get_sub_epoch_dataset(files_subepoch: List[str]) -> Dict[str, Any]:
          from returnn.util.file_cache import CachedFile
          return {
            "class": "HDFDataset",
            "files": [CachedFile(fn) for fn in files_subepoch],
          }

        train = {
          "class": "ConcatFilesDataset",
          "files": [
            "/nfs/big_data_1.hdf",
            ...
          ],  # M files
          "get_sub_epoch_dataset": get_sub_epoch_dataset,
          "partition_epoch": P,  # P << M
        }

    Instead of a plain list of strings, you can also provide a list of any nested structures to keep
    multimodal data together::

        def get_sub_epoch_dataset(files_subepoch: List[Tuple[str, str]]) -> Dict[str, Any]:
          from returnn.util.file_cache import CachedFile

          alignments, features = tuple(zip(*files_subepoch)) # transpose

          return {
            "class": "MetaDataset",
            "data_map": {"classes": ("alignments", "data"), "data": ("features", "data")},
            "datasets": {
              "alignments": {
                "class": "HDFDataset",
                "files": alignments,
                "seq_ordering": "random",
              },
              "features": {
                "class": "HDFDataset",
                "files": features,
              },
            },
            "seq_order_control_dataset": "alignments",
          }

        train = {
          "class": "ConcatFilesDataset",
          "files": [
            ("/nfs/alignment_1.hdf", "/nfs/features_1.hdf"),
            ...
          ],  # M entries
          "get_sub_epoch_dataset": get_sub_epoch_dataset,
          "partition_epoch": P,  # P << M
        }

    In this case the file sizes for sub epoch distribution are summed up per list entry
    by iterating over the structure leaves.

    For some discussion, see https://github.com/rwth-i6/returnn/issues/1519 and
    https://github.com/rwth-i6/returnn/issues/1524.
    """

    def __init__(
        self,
        *,
        files: List[FileTree],
        get_sub_epoch_dataset: Callable[[List[FileTree]], Dict[str, Any]],
        preload_next_n_sub_epochs: int = 1,
        buffer_size: int = 1,
        file_cache_opts: Optional[Dict[str, Any]] = None,
        _meta_info_cache: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        :param files: the files to shuffle over, can also be a list of arbitrarily nested python objects
            to keep associated heterogeneous data together
        :param get_sub_epoch_dataset: callable which returns a dataset dict for a given subset of files
        :param preload_next_n_sub_epochs: how many sub epoch datasets to preload
        :param buffer_size: buffer size for each worker, amount of seqs to prefetch
        :param _meta_info_cache: for internal use
        """
        super().__init__(**kwargs)
        self.files = files
        self.get_sub_epoch_dataset = get_sub_epoch_dataset
        assert preload_next_n_sub_epochs >= 0
        self.preload_next_n_sub_epochs = preload_next_n_sub_epochs
        self.buffer_size = buffer_size
        self.file_cache_opts = file_cache_opts or {}
        self._file_sizes: Optional[Dict[str, int]] = None  # key -> size. for equal distribution across sub epochs
        self._data_keys: Optional[List[str]] = None
        self._num_seqs: Optional[int] = None

        self._file_cache: Optional[_FileCacheProc] = None
        self._workers: Dict[int, _WorkerProcParent] = {}  # epoch -> worker
        self._files_order_cache: Dict[int, List[List[FileTree]]] = {}  # full epoch (0-indexed) -> files order

        if _meta_info_cache:
            # This allows to skip the lazy init in self.initialize().
            # This can be used when pickling/deepcopying an instance of this dataset,
            # see Dataset.__reduce__.
            self.num_inputs = _meta_info_cache["num_inputs"]
            self.num_outputs = _meta_info_cache["num_outputs"]
            self.labels = _meta_info_cache["labels"]
            self._data_keys = _meta_info_cache["data_keys"]
            self._file_sizes = _meta_info_cache["file_sizes"]

        if len(files) < self.partition_epoch:
            raise ValueError(f"{self}: len(files) {len(files)} < partition_epoch {self.partition_epoch}")

    def initialize(self):
        """init"""
        self._lazy_init_num_outputs()
        super().initialize()

    @property
    def _meta_info_cache(self):
        if not self.num_outputs:
            return None
        return {
            "num_inputs": self.num_inputs,
            "num_outputs": self.num_outputs,
            "labels": self.labels,
            "data_keys": self._data_keys,
            "file_sizes": self._file_sizes,
        }

    def _lazy_init_num_outputs(self):
        if self.num_outputs:
            return
        # First, we need to know the num_inputs, num_outputs, total_num_seqs, labels.
        # Init the dataset with the first file.
        dataset_dict, exit_hook = self._get_sub_dataset_dict(files=[self.files[0]])
        try:
            dataset = init_dataset(dataset_dict, extra_kwargs={"seq_ordering": "default"})
            self.num_inputs = dataset.num_inputs
            self.num_outputs = dataset.num_outputs
            self.labels = dataset.labels
            self._data_keys = dataset.get_data_keys()
        finally:
            exit_hook()

    def _lazy_init_file_sizes(self):
        import tree

        if self._file_sizes:
            return
        self._file_sizes = {
            _get_key_for_file_tree(t): sum((os.path.getsize(fn) for fn in tree.flatten(t)), 0) for t in self.files
        }

    def _lazy_init_file_cache_proc(self):
        if self._file_cache:
            return
        self._file_cache = _FileCacheProc(
            name=f"{self.__class__.__name__} {self.name}", file_cache_opts=self.file_cache_opts
        )

    def __del__(self):
        if self._file_cache:
            try_run(self._file_cache.exit, kwargs={"join": False})
            self._file_cache = None
        for k, worker in self._workers.items():
            try_run(worker.exit, kwargs={"join": False})
        self._workers.clear()

    def init_seq_order(self, epoch: Optional[int] = None, seq_list=None, seq_order=None) -> bool:
        """
        :param epoch:
        :param seq_list:
        :param seq_order:
        :return: whether the order changed (True is always safe to return)
        """
        super().init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)
        if seq_list is not None:
            raise Exception(f"{self}: seq_list not supported")
        if seq_order is not None:
            raise Exception(f"{self}: seq_order not supported")
        if epoch is None:
            self._num_seqs = 0
            return True

        self._lazy_init_file_sizes()
        self._lazy_init_file_cache_proc()

        # Cleanup and fill _files_order_cache.
        full_epoch_0idx = (epoch - 1) // self.partition_epoch
        for k in list(self._files_order_cache.keys()):
            if k < full_epoch_0idx:
                del self._files_order_cache[k]
        for ep_ in range(epoch, epoch + self.preload_next_n_sub_epochs + 1):
            full_epoch_0idx_ = (ep_ - 1) // self.partition_epoch
            if full_epoch_0idx_ in self._files_order_cache:
                continue
            if self.seq_ordering == "default":
                files_order_flat = self.files
            elif self.seq_ordering == "random":
                rnd_seed = self._get_random_seed_for_epoch(full_epoch_0idx_ * self.partition_epoch + 1)
                random_generator = numpy.random.RandomState(rnd_seed)
                files_order_flat = list(self.files)
                random_generator.shuffle(files_order_flat)
            else:
                raise ValueError(f"{self}: seq_ordering {self.seq_ordering!r} not supported")
            self._files_order_cache[full_epoch_0idx_] = self._get_files_per_sub_epochs(
                partition_epoch=self.partition_epoch, file_sizes=self._file_sizes, files_order=files_order_flat
            )

        # Cleanup and fill _workers.
        for k, worker in list(self._workers.items()):
            if k < epoch:
                worker.exit()
                del self._workers[k]
        for ep_ in range(epoch, epoch + self.preload_next_n_sub_epochs + 1):
            if ep_ in self._workers:
                continue
            full_epoch_0idx_ = (ep_ - 1) // self.partition_epoch
            files_order: List[List[FileTree]] = self._files_order_cache[full_epoch_0idx_]
            files_for_subep = files_order[(ep_ - 1) % self.partition_epoch]
            print(f"{self}: using files for epoch {ep_}: {files_for_subep}", file=log.v4)
            dataset_dict, exit_hook = self._get_sub_dataset_dict(files=files_for_subep)
            worker = _WorkerProcParent(
                name=f"{self.__class__.__name__} {self.name} ep {epoch}",
                epoch=ep_,
                full_epoch_0idx=full_epoch_0idx_,
                dataset_dict=dataset_dict,
                buffer_size=self.buffer_size,
                exit_hook=exit_hook,
            )
            self._workers[ep_] = worker

        self._num_seqs = self._workers[epoch].get_num_seqs()
        return True

    def _get_sub_dataset_dict(self, files: List[FileTree]) -> Tuple[Dict[str, Any], _FileCacheExitHook]:
        dataset_dict = self.get_sub_epoch_dataset(files)
        if "random_seed_offset" not in dataset_dict:
            dataset_dict["random_seed_offset"] = self.random_seed_offset
        if dataset_dict.get("partition_epoch", 1) != 1:
            raise ValueError(f"{self}: sub dataset should not have partition_epoch, got: {dataset_dict}")
        if "seq_ordering" not in dataset_dict and "seq_order_control_dataset" not in dataset_dict:
            raise ValueError(
                f"{self}: sub dataset should have explicit seq_ordering "
                "(or seq_order_control_dataset for MetaDataset), got: {dataset_dict}"
            )
        self._lazy_init_file_cache_proc()
        dataset_dict, exit_hook = self._file_cache.handle_cached_files_in_config(dataset_dict)
        return dataset_dict, exit_hook

    @staticmethod
    def _get_files_per_sub_epochs(
        *, partition_epoch: int, file_sizes: Dict[str, int], files_order: Sequence[FileTree]
    ) -> List[List[FileTree]]:
        total_size = sum(file_sizes.values())
        avg_size_per_sub_epoch = total_size / partition_epoch
        # Now evenly distribute the files over the sub epochs.
        # Note that many one-pass variants of algorithms to evenly distribute
        # can end up with some empty sub epochs,
        # so we need to make sure that this is not the case.
        # E.g. consider the seqs of size [1,1,78,120] and partition_epoch=4.
        # That has avg size per sub epoch 50.
        # Some simple algorithms could end up with the sub epochs
        # [[1,1], [78], [120], []] or [[1,1,78], [120], [], []].
        # Or consider [5,5]+[10]*7, partition_epoch=5, which has avg size 16.
        # A simple algorithm could end up with [[5,5,10], [10,10], [10,10], [10,10], []].
        # See test_ConcatFilesDataset_get_files_per_sub_epochs for some test cases.
        assert len(files_order) >= partition_epoch
        files_per_sub_epochs = [[] for _ in range(partition_epoch)]
        assert len(files_per_sub_epochs) == partition_epoch
        sub_epoch_idx = 0
        size_taken = 0
        for i, f_tree in enumerate(files_order):
            size = file_sizes[_get_key_for_file_tree(f_tree)]
            num_remaining = len(files_order) - i
            if num_remaining <= partition_epoch - sub_epoch_idx - 1:
                # All remaining sub epochs must be filled.
                assert files_per_sub_epochs[sub_epoch_idx]
                sub_epoch_idx += 1
                files_per_sub_epochs[sub_epoch_idx].append(f_tree)
                size_taken = size
                continue
            if sub_epoch_idx == partition_epoch - 1:
                # We are done. Just add the rest to the last sub epoch.
                files_per_sub_epochs[sub_epoch_idx].append(f_tree)
                size_taken += size
                continue
            assert size_taken <= avg_size_per_sub_epoch
            if size_taken + size <= avg_size_per_sub_epoch:
                files_per_sub_epochs[sub_epoch_idx].append(f_tree)
                size_taken += size
                continue
            # We should increase the sub epoch index.
            # We need to decide where to add this file, to the current or the next sub epoch.
            if not files_per_sub_epochs[sub_epoch_idx] or (
                # Better to add this file to the current sub epoch?
                abs((size_taken + size) - avg_size_per_sub_epoch)
                <= abs(size_taken - avg_size_per_sub_epoch)
            ):
                files_per_sub_epochs[sub_epoch_idx].append(f_tree)
                size_taken = 0
            else:
                files_per_sub_epochs[sub_epoch_idx + 1].append(f_tree)
                size_taken = size
            sub_epoch_idx += 1
        assert all(files_per_sub_epochs)
        return files_per_sub_epochs

    def _collect_single_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        if seq_idx >= self._num_seqs:
            return None
        seq = self._workers[self.epoch].get_data_seq(seq_idx)
        if not seq:
            return None
        seq.seq_idx = seq_idx
        return seq

    def have_seqs(self) -> bool:
        """have seqs"""
        return bool(self.files)

    @property
    def num_seqs(self) -> int:
        """num seqs"""
        return self._num_seqs

    def finish_epoch(self, *, free_resources: bool = False):
        """finish epoch"""
        super().finish_epoch(free_resources=free_resources)
        if free_resources:
            for worker in self._workers.values():
                worker.exit()
            self._workers.clear()
            self._files_order_cache.clear()
            if self._file_cache:
                self._file_cache.exit()
                self._file_cache = None
        else:
            if self.epoch in self._workers:
                worker = self._workers.pop(self.epoch)
                worker.exit()

    def get_data_keys(self) -> List[str]:
        """data keys"""
        if self._data_keys is None:
            self._lazy_init_num_outputs()
        return self._data_keys


def _get_key_for_file_tree(t: FileTree) -> str:
    """generates a deterministic key given a file tree"""
    import tree

    return ":".join(tree.flatten(t))


class _WorkerProcParent:
    def __init__(
        self,
        *,
        name: str,
        epoch: int,
        full_epoch_0idx: int,
        dataset_dict: Dict[str, Any],
        buffer_size: int,
        exit_hook: Optional[Callable[[], None]] = None,
    ):
        self.epoch = epoch
        self.full_epoch_0idx = full_epoch_0idx
        self.dataset_dict = dataset_dict
        self.exit_hook = exit_hook

        parent_conn, child_conn = _mp.Pipe()
        self.parent_conn: mpConnection = parent_conn

        self.worker_proc = _mp.Process(
            name=f"{name} worker ep {epoch}",
            target=_worker_proc_loop,
            args=(epoch, buffer_size, dataset_dict, child_conn),
            daemon=True,
        )
        self.worker_proc.start()
        # Make sure the child connection is closed here.
        # It stays open in the child, until the child dies.
        # When that happens, now any consecutive read on the pipe
        # should yield an exception -- which is what we want,
        # otherwise it would just hang.
        child_conn.close()

        # This worker operates on one sub epoch.
        # The data used here is only for this sub epoch.
        # Without shuffling, the next time we would use this data is in the next full epoch.
        # Thus, use the full epoch as the epoch here.
        # Make sure this init is async - we should not wait for this in the main process when it is not used yet.
        self.parent_conn.send(("init_seq_order", {"epoch": self.full_epoch_0idx + 1}))
        self._num_seqs: Optional[int] = None

    def _lazy_wait_for_init_seq_order(self):
        if self._num_seqs is not None:
            return
        msg, num_seqs = self.parent_conn.recv()
        assert msg == "num_seqs" and isinstance(num_seqs, int)
        self._num_seqs = num_seqs

    def get_num_seqs(self) -> int:
        """num seqs for this sub epoch"""
        self._lazy_wait_for_init_seq_order()
        return self._num_seqs

    def get_data_seq(self, seq_idx: int) -> Optional[DatasetSeq]:
        """get data seq"""
        self._lazy_wait_for_init_seq_order()
        self.parent_conn.send(("get_data_seq", {"seq_idx": seq_idx}))
        msg, data = self.parent_conn.recv()
        assert msg == "data_seq"
        return data

    def exit(self, *, join: bool = True):
        """exit"""
        self._lazy_wait_for_init_seq_order()
        self.parent_conn.send(("exit", {}))
        if join:
            self.worker_proc.join()
        if self.exit_hook:
            self.exit_hook()
            self.exit_hook = None

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.exit(join=False)
        except Exception:
            pass
        else:
            try_run(self.worker_proc.join)


def _worker_proc_loop(
    epoch: int,
    buffer_size: int,
    dataset_dict: Dict[str, Any],
    parent_conn: mpConnection,
):
    if sys.platform == "linux":
        with open("/proc/self/comm", "w") as f:
            f.write(f"CFD worker {epoch}")

    assert isinstance(epoch, int) and isinstance(buffer_size, int)
    assert isinstance(dataset_dict, dict)
    assert isinstance(parent_conn, mpConnection)

    dataset = init_dataset(dataset_dict)

    got_init_seq_order = False
    cache: List[DatasetSeq] = []
    next_seq_idx = 0

    # noinspection PyShadowingNames
    def _add_to_cache():
        nonlocal next_seq_idx
        if len(cache) >= buffer_size:
            return False
        if not dataset.is_less_than_num_seqs(next_seq_idx):
            return False
        dataset.load_seqs(next_seq_idx, next_seq_idx + 1)
        seq_tag = dataset.get_tag(next_seq_idx)
        features = {data_key: dataset.get_data(next_seq_idx, data_key) for data_key in dataset.get_data_keys()}
        res = DatasetSeq(seq_idx=next_seq_idx, seq_tag=seq_tag, features=features)
        cache.append(res)
        next_seq_idx += 1
        return True

    # noinspection PyShadowingNames
    def _get_from_cache(seq_idx: int) -> Optional[DatasetSeq]:
        if not cache:
            return None
        if seq_idx > cache[-1].seq_idx:
            return None
        for seq in cache:
            assert seq.seq_idx <= seq_idx
            if seq.seq_idx == seq_idx:
                return seq
        assert False

    # noinspection PyShadowingNames
    def _get(seq_idx: int) -> Optional[DatasetSeq]:
        if cache and seq_idx < cache[0].seq_idx:
            raise Exception(
                f"requested seq idx {seq_idx} is smaller than cache start {cache[0].seq_idx}, cannot go backwards"
            )
        res = _get_from_cache(seq_idx)
        if res:
            return res
        if not dataset.is_less_than_num_seqs(seq_idx):
            return None
        assert next_seq_idx <= seq_idx
        while True:
            if not _add_to_cache():
                raise Exception(
                    f"buffer too small, requested seq idx {seq_idx},"
                    f" cache starts at {cache[0].seq_idx if cache else None}"
                )
            assert cache[-1].seq_idx <= seq_idx
            if cache[-1].seq_idx == seq_idx:
                return cache[-1]

    try:
        while True:
            if got_init_seq_order:
                while not parent_conn.poll():
                    if not _add_to_cache():
                        break
            msg, kwargs = parent_conn.recv()
            if msg == "exit":
                break
            elif msg == "get_data_seq":
                seq_idx = kwargs["seq_idx"]
                while cache and cache[0].seq_idx < seq_idx:
                    cache.pop(0)
                res = _get(seq_idx)
                parent_conn.send(("data_seq", res))
            elif msg == "init_seq_order":
                # We are responsible to get the seq order and distrib it to all the other workers.
                dataset.init_seq_order(**kwargs)
                parent_conn.send(("num_seqs", dataset.num_seqs))
                got_init_seq_order = True
                next_seq_idx = 0
                cache[:] = []
            else:
                raise Exception(f"unknown msg {msg!r}")
    except KeyboardInterrupt:  # when parent dies
        pass


class _FileCacheProc:
    def __init__(self, *, name: str, file_cache_opts: Dict[str, Any]):
        self.file_cache_opts = file_cache_opts

        parent_conn, child_conn = _mp.Pipe()
        self.parent_conn: mpConnection = parent_conn

        self.proc = _mp.Process(
            name=f"{name} file cache",
            target=_file_cache_proc_loop,
            args=(file_cache_opts, child_conn),
            daemon=True,
        )
        self.proc.start()
        # Make sure the child connection is closed here.
        # It stays open in the child, until the child dies.
        # When that happens, now any consecutive read on the pipe
        # should yield an exception -- which is what we want,
        # otherwise it would just hang.
        child_conn.close()

    def handle_cached_files_in_config(self, config: Dict[str, Any]) -> Tuple[Dict[str, Any], _FileCacheExitHook]:
        """:func:`FileCache.handle_cached_files_in_config`"""
        self.parent_conn.send(("handle_cached_files_in_config", {"config": config}))
        msg, (res, files) = self.parent_conn.recv()
        assert msg == "config_and_cached_files" and isinstance(config, dict) and isinstance(files, list)
        return res, _FileCacheExitHook(self, files)

    def release_files(self, filenames: Collection[str]):
        """:func:`FileCache.release_files`"""
        self.parent_conn.send(("release_files", {"filenames": filenames}))

    def exit(self, *, join: bool = True):
        """exit"""
        if self.proc.exitcode is not None:  # already stopped?
            return
        self.parent_conn.send(("exit", {}))
        if join:
            self.proc.join()

    def __del__(self):
        # noinspection PyBroadException
        try:
            self.exit(join=False)
        except Exception:
            pass
        else:
            try_run(self.proc.join)


class _FileCacheExitHook:
    def __init__(self, file_cache_proc: _FileCacheProc, cached_files: List[str]):
        self.file_cache_proc = file_cache_proc
        self.cached_files = cached_files

    def __call__(self):
        if self.file_cache_proc.proc.exitcode is not None:
            return
        self.file_cache_proc.release_files(self.cached_files)


def _file_cache_proc_loop(
    file_cache_opts: Dict[str, Any],
    parent_conn: mpConnection,
):
    if sys.platform == "linux":
        with open("/proc/self/comm", "w") as f:
            f.write(f"File cache")

    assert isinstance(file_cache_opts, dict)
    assert isinstance(parent_conn, mpConnection)

    from returnn.util.file_cache import FileCache

    file_cache = FileCache(**file_cache_opts)

    try:
        while True:
            msg, kwargs = parent_conn.recv()
            if msg == "exit":
                break
            elif msg == "handle_cached_files_in_config":
                res, files = file_cache.handle_cached_files_in_config(**kwargs)
                parent_conn.send(("config_and_cached_files", (res, files)))
            elif msg == "release_files":
                file_cache.release_files(**kwargs)
            else:
                raise Exception(f"unknown msg {msg!r}")
    except KeyboardInterrupt:  # when parent dies
        pass
